#include "sd.h"

// Private Prototypes
static euint8 SD_WriteCommandR1(euint8 command, euint32 arg);
euint8 SD_WriteCommandR2(euint8 command, euint32 arg);
static euint8 SD_ReadMedia(void);

/*
 * Returns:		-1 Timeout while resetting SD-Card.
 *				-2 Timeout during initialization
 *				-3 Error while trying to set blocklength (512)
 * 				 0 Success
 */
esint8 SD_Init(void)
{
	euint8 response;
	euint16 i;

	/* Configure AN0..AN11 to be digital IO-Pins */
	ADCON1 = 0x0F;
	/* Make the corresponding chip select pin an digital output */
	SD_CS_DIR = 0;
    SD_Disable();

	/* Let the card power up. Wait 1M CLK cycles @10MHz -> 100ms */
	Delay10KTCYx(100);

    /* card needs 74 cycles minimum to start up */
    for(i = 0; i < 0x0f; ++i)
    {
        /* wait 8 clock cycles sending dummy data*/
        WriteSPI(0xff);
    }
	#if SD_DEBUG
	printf((rom char*)"SD Card powered up.\n\r");
	#endif

    /* The SD-Card will enter the SPI mode if the CS-Signal 
	is asserted during the reception of the reset command (CMD0)*/
    SD_Enable();

	/* Wait 1ms */
	Delay10KTCYx(1);
    
    for(i = 0; ; ++i)
    {
        response = SD_WriteCommandR1(CMD_GO_IDLE_STATE, 0);
		#if SD_DEBUG_LOWLVL
		printf((rom char*)"Send CMD0. Received: %#x\n\r", response);
		#endif
        if(response == (1 << R1_IDLE_STATE))
            break;

        if(i == 0x100)
        {
            SD_Disable();
			#if SD_DEBUG
			printf((rom char*)"Timeout. Failed to reset the SD-Card.\n\r");
			#endif
            return(-1);
        }
    }
	
	#if SD_DEBUG
	printf((rom char*)"Reseted SD-Card to idle state. Response: %#x\n\r", response);
	#endif    

    /* Activate the Card initialization process. The host has to poll the SD-Card
 	using the CMD1 until the idle state is unsetted.*/
    for(i = 0; ; ++i)
    {
        response = SD_WriteCommandR1(CMD_SEND_OP_COND, 0);
		#if SD_DEBUG_LOWLVL
		printf((rom char*)"Send CMD1. Received: 0x%x\n\r", response);
		#endif
        if(!response) // received 0x00
            break;

        if(i == 0x2FF)
        {
			#if SD_DEBUG
			printf((rom char*)"Timeout with CMD1. Try again via ACMD41.\n\r");
			#endif
			/* Try alternative initialization for SDC */
			while(i){
				if(SD_WriteCommandR1(CMD_APP_CMD, 0) != 0xFE) continue;
				if(!(response = SD_WriteCommandR1(ACMD_SEND_OP_COND, 0)))
					break;
			}
			if(!response)
				break;
				
            SD_Disable();
			#if SD_DEBUG
			printf((rom char*)"Timeout. Failed to init the SD-Card.\n\r");
			#endif
            return(-2);
        }
    }
	#if SD_DEBUG
	printf((rom char*)"Iinitialized SD-Card. Response: %#x\n\r", response);
	#endif

    SD_Disable();
	
	return 0;
}

/*
 * Returns:		resonse - Card response to the given cmd
 */
static euint8 SD_WriteCommandR1(euint8 command, euint32 arg)
{
	euint8 i;
	euint8 response;
	euint8 savedSD_CS = SD_CS;

	//WriteSPI(0xff); 	// Dummy delay 8 Bit
	
	SD_Enable();

    /* send command via SPI */
    WriteSPI(0x40 | command);
    WriteSPI((arg >> 24) & 0xff);
    WriteSPI((arg >> 16) & 0xff);
    WriteSPI((arg >> 8) & 0xff);
    WriteSPI(arg & 0xff);
    WriteSPI(command == CMD_GO_IDLE_STATE ? 0x95 : 0xff); // Calculate CRC only on reset
    
    /* receive response */
    for(i = 0; i < 20; ++i)
    {
        response = SD_ReadMedia();
        if(response != 0xff)
            break;
    }
	
	SD_Disable();
	
	// Following any command, the SD Card needs 8 clocks to finish up its work.
	// (from SanDisk SD Card Product Manual v1.9 section 5.1.8)
	WriteSPI(0xff);
	
	SD_CS = savedSD_CS;

    return response;
}

/*
 * Returns:		resonse - Card response to the given cmd
 */
euint8 SD_WriteCommandR2(euint8 command, euint32 arg)
{
	euint8 i;
    euint16 response;
	euint8 savedSD_CS = SD_CS;    
	
	//WriteSPI(0xff); 	// Dummy delay 8 Bit
	
	SD_Enable();

    /* send command via SPI */
    WriteSPI(0x40 | command);
    WriteSPI((arg >> 24) & 0xff);
    WriteSPI((arg >> 16) & 0xff);
    WriteSPI((arg >> 8) & 0xff);
    WriteSPI(arg & 0xff);
    WriteSPI(command == CMD_GO_IDLE_STATE ? 0x95 : 0xff); // Calculate CRC only on reset
    
    /* receive response */
    for(i = 0; i < 20; ++i)
    {
        response = SD_ReadMedia();
        if(response != 0xff)
            break;
    }
    response <<= 8;
    response |= SD_ReadMedia();

	SD_Disable();
	
	// Following any command, the SD Card needs 8 clocks to finish up its work.
	// (from SanDisk SD Card Product Manual v1.9 section 5.1.8)
	WriteSPI(0xff);
	
	SD_CS = savedSD_CS;

    return response;
}

/*
 * Returns:		-1 Read cmd was not accepted
 *				-2 Timeout while waiting for data response
 *				 0 Success
 */
esint8 SD_ReadBlock(euint32 addr, eint8* buf)
{
	euint16 i, timeout, response;
	euint8 status;

	SD_Enable();

	// Send the read command
	if(response = SD_WriteCommandR1(CMD_READ_SINGLE_BLOCK, addr)){
		#if SD_DEBUG
		printf((rom char*)"Read command was not accepted. Response: %#x\n\r", response);
		#endif
		SD_Disable();
		return -1;
	}
	
	#if SD_DEBUG
	printf((rom char*)"Read command accepted. Received: %#x\n\r", status);
	#endif
	
	timeout = 0x2FF;

	// Now wait for the "Start Block" token	(0xFE)
	// (see SanDisk SD Card Product Manual v1.9 section 5.2.4. Data Tokens)
	do {
		status = SD_ReadMedia();
		#if SD_DEBUG_LOWLVL
		printf((rom char*)"Waiting for 0xFE. Received: %#x\n\r", status);
		#endif
	} while(status != 0xFE && timeout--);

	if(!timeout){ //0x00 -> timed out.
		SD_Disable();
		#if SD_DEBUG
		printf((rom char*)"Timeout while waiting for data start token\n\r");
		#endif
		return -2;
	}
	
	// Read off all the bytes in the block
	for(i = 0; i < SD_BLOCK_SIZE; i++){
		buf[i] = SD_ReadMedia();
	}
	
	// Read CRC bytes
	status = SD_ReadMedia();
	status = SD_ReadMedia();

	SD_Disable();

	// Following a read transaction, the SD Card needs 8 clocks after the end 
	// bit of the last data block to finish up its work.
	// (from SanDisk SD Card Product Manual v1.9 section 5.1.8)
	WriteSPI(0xFF); 

	return(0);
}

/*
 * Returns:		-1 Card is writeprotected
 * 				-2 Write cmd was not accepted
 *				-3 Timeout while processing write cmd
 *				 0 Success
 */
esint8 SD_WriteBlock(euint32 addr, eint8* buf)
{
	euint16 i, response, timeout;
	
	/*if(SD_WriteProtect())
		return -1;*/
	
	SD_Enable();
	if(response = SD_WriteCommandR1(CMD_WRITE_SINGLE_BLOCK, addr))
	{
		SD_Disable();
		#if SD_DEBUG
		printf((rom char*)"Did not accept write CMD.Response: %#x\n\r",response);
		#endif
		return -2;
	}
	
	#if SD_DEBUG
	printf((rom char*)"Write CMD accepted. Received: %#x\n\r",response);
	#endif

	for(i = 0; i < 0xff; i++)	// Delay befor sending the data
		WriteSPI(0xff);
	
	
	/* Write Send byte */
	WriteSPI(0xfe);
	
	/* Send data */
	for(i=0; i < SD_BLOCK_SIZE; i++)
	{
		WriteSPI(buf[i]);
	}
	
	/* Send checksum */
	WriteSPI(0xff);
	WriteSPI(0xff);
	
	timeout = 0xfff;
	/* Wait while card is busy */
	while((SD_ReadMedia() != 0xff) && timeout--);

	if(!timeout){
		SD_Disable();
		#if SD_DEBUG
		printf((rom char*)"Timed out while processing write CMD\n\r");
		#endif
		return(-3);
	}
	
	SD_Disable();
		
	// Following a read transaction, the SD Card needs 8 clocks after the end 
	// bit of the last data block to finish up its work.
	// (from SanDisk SD Card Product Manual v1.9 section 5.1.8)
	WriteSPI(0xFF); 

	return(0);	
}

esint8 SD_ReadCSD(eint8* buf)
{
	euint16 i, timeout, response;
	euint8 status;

	SD_Enable();

	// Send the read command
	if(response = SD_WriteCommandR1(CMD_SEND_CSD, 0x00)){
		SD_Disable();
		#if SD_DEBUG
		printf((rom char*)"Read CSD command was not accepted. Response: %#x\n\r", response);
		#endif
		return -1;
	}
	
	timeout = 0x2FF;

	// Now wait for the "Start Block" token	(0xFE)
	// (see SanDisk SD Card Product Manual v1.9 section 5.2.4. Data Tokens)
	do {
		status = SD_ReadMedia();
	} while(status != 0xFE && timeout--);

	if(!timeout){ //0x00 -> timed out.
		SD_Disable();
		#if SD_DEBUG 
		printf((rom char*)"Timeout while waiting for data start token\n\r");
		#endif
		return -2;
	}
	
	// Read off all the bytes in the block
	for(i = 0; i < CSD_SIZE; i++){
		buf[i] = SD_ReadMedia();
	}

	/* Read CRC. Specs 5.1.6 */
	SD_ReadMedia();
	SD_ReadMedia();


	SD_Disable();

	// Following a read transaction, the SD Card needs 8 clocks after the end 
	// bit of the last data block to finish up its work. The status of the CS-Signal
	// is irrelevant.
	// (from SanDisk SD Card Product Manual v1.9 section 5.1.8)
	WriteSPI(0xFF); 

	return(0);
}



/* Because of the SD Card Specifications I am unable to use the MCC18 ReadSPI()-Function.
   The SDO has to be held high during data reception. */
static euint8 SD_ReadMedia(void)
{
    SSPBUF = 0xFF;                              //Data Out - Logic ones
    while(!SSPSTATbits.BF);                     //Wait until cycle complete
    return(SSPBUF);                             //Return with byte read 
}//end ReadMedia

